from pydantic import BaseModel
from typing import Optional

class FaktsEndpoint(BaseModel):
    base_url = "http://localhost:8000/f/"
    name: str = "Helper"
    description: Optional[str] 


class Discovery(BaseModel):
    """Discovery is the abstract base class for discovery mechanisms

    A discovery mechanism is a way to find a Fakts endpoint
    that can be used to retrieve the configuration.

    This class provides an asynchronous interface, as the discovery can
    envolve lenghty operations such as network requests or waiting for
    user input.
    """

    async def discover(self, force_refresh=False) -> FaktsEndpoint:
        raise NotImplementedError("Discovery needs to implement this function")



import aiohttp
from fakts.grants.remote.base import RemoteGrant
from fakts.discovery.base import FaktsEndpoint


class RetrieveException(Exception):
    pass


class RetrieveGrant(RemoteGrant):
    """Retrieve Grant

    A retrieve grant is a remote grant can be used to retrieve a token and a configuration from a fakts server, by claiming to be an already
    registed public application on the fakts server. Public applications are applications that are not able to keep a secret, and therefore
    need users to explicitly grant them access to their data. YOu need to also provide a redirect_uri that matches the one that is registered
    on the fakts server.

    """

    redirect_uri: str

    async def ademand(self, endpoint: FaktsEndpoint) -> str:
        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self.ssl_context)
        ) as session:
            print(f"{endpoint.base_url}retrieve/")
            async with session.post(
                f"{endpoint.base_url}retrieve/",
                json={
                    "version": self.version,
                    "identifier": self.identifier,
                    "redirect_uri": self.redirect_uri,
                },
            ) as resp:
                data = await resp.json()

                if resp.status == 200:
                    data = await resp.json()
                    if not "status" in data:
                        raise RetrieveException("Malformed Answer")

                    status = data["status"]
                    if status == "error":
                        raise RetrieveException(data["message"])
                    if status == "granted":
                        return data["token"]

                    raise RetrieveException(f"Unexpected status: {status}")
                else:
                    raise Exception("Error! Coud not claim this app on this endpoint")



class FaktsError(Exception):
    """Base class for all Fakts errors"""



class NoFaktsFound(FaktsError):
    """Raised when no fakts instance is found in the current context"""



class GroupNotFound(FaktsError):
    """Raised when a group is not found in the configuration"""




from rich import get_console
from fakts.discovery.beacon import retrieve_bindings, EndpointBeacon
from fakts.discovery.base import FaktsEndpoint
from rich.prompt import Prompt
import argparse


def main(name: str = None, url: str = None):
    """Advertises the given endpoint on the interface specified by the user


    Args:
        name (str, optional): The name of the faktsendpoint. Defaults to None (user can specify it)
        url (str, optional): The url of the faktsendpont. Defaults to None (user can specify it)
    """

    if not name:
        name = Prompt.ask(
            "How do you want this beacon to be advertisted as?", default="Arkitekt"
        )

    get_console().print("Which Interface should be used for broadcasting?")
    bindings = retrieve_bindings()
    for i, binding in enumerate(bindings):
        get_console().print(
            f"[{i}] : Use interface {binding.interface}: {binding.bind_addr} advertising to {binding.broadcast_addr}"
        )

    bind_index = Prompt.ask(
        "Which Binding do you want to use?",
        default=1,
        choices=[str(i) for i in range(len(bindings))],
    )

    if not url:
        url = Prompt.ask(
            "Which Setup Uri do you want to advertise?",
            default=f"http://{bindings[int(bind_index)].bind_addr}:8000/f/",
        )

    with EndpointBeacon(
        advertised_endpoints=[FaktsEndpoint(base_url=url, name=name)],
        binding=bindings[int(bind_index)],
    ) as beacon:
        beacon.run()


def entrypoint():
    parser = argparse.ArgumentParser(description="Say hello")
    parser.add_argument("--url", type=str, help="The Name of this script")
    parser.add_argument("--name", type=bool, help="Do you want to refresh")
    args = parser.parse_args()

    main(name=args.name, url=args.url)


if __name__ == "__main__":
    entrypoint()



from pydantic import Field
from qtpy.QtCore import Signal
from fakts.discovery.advertised import AdvertisedDiscovery
from fakts.discovery.base import FaktsEndpoint
from qtpy import QtWidgets, QtCore, QtGui
import asyncio
import logging
from koil.qt import QtCoro, QtFuture
import ssl
import certifi
import aiohttp
logger = logging.getLogger(__name__)


class SelfScanWidget(QtWidgets.QWidget):
    user_endpoint = Signal(FaktsEndpoint)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.scanlayout = QtWidgets.QHBoxLayout()
        self.lineEdit = QtWidgets.QLineEdit()
        self.addButton = QtWidgets.QPushButton("Scan")

        self.scanlayout.addWidget(self.lineEdit)
        self.scanlayout.addWidget(self.addButton)
        self.addButton.clicked.connect(self.on_add)
        self.setLayout(self.scanlayout)

    def on_add(self):
        host = self.lineEdit.text()
        endpoint = FaktsEndpoint(base_url=host, name="Self Added")
        self.user_endpoint.emit(endpoint)


class FaktsEndpointWidget(QtWidgets.QWidget):

    accept_clicked = QtCore.Signal(FaktsEndpoint)

    def __init__(self, endpoint: FaktsEndpoint, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.endpoint = endpoint

        self.user_label = QtWidgets.QLabel(endpoint.name)
        self.user_label.setStyleSheet("font-size: 20px;")
        self.endpoint_label = QtWidgets.QLabel(endpoint.base_url)
        self.endpoint_label.setStyleSheet("font-size: 10px;")
        self.user_label.mousePressEvent = self.on_clicked

        self.hlayout = QtWidgets.QVBoxLayout()
        self.setLayout(self.hlayout)
        self.hlayout.addWidget(self.user_label)
        self.hlayout.addWidget(self.endpoint_label)

    def on_clicked(self, event):
        self.accept_clicked.emit(self.endpoint)




class SelectBeaconWidget(QtWidgets.QDialog):
    new_endpoint = Signal(FaktsEndpoint)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setWindowTitle("Search Endpoints...")
        self.show_coro = QtCoro(self.show_me, autoresolve=True)
        self.hide_coro = QtCoro(self.hide, autoresolve=True)

        self.select_endpoint = QtCoro(self.demand_selection_of_endpoint)
        self.select_endpoint_future = None

        self.new_endpoint.connect(self.on_new_endpoint)

        self.endpoints = []

        self.endpointLayout = QtWidgets.QVBoxLayout()

        self.scanWidget = SelfScanWidget()
        self.scanWidget.user_endpoint.connect(self.on_new_endpoint)

        QBtn = QtWidgets.QDialogButtonBox.Cancel
        self.buttonBox = QtWidgets.QDialogButtonBox(QBtn)
        self.buttonBox.rejected.connect(self.on_reject)

        self.wlayout = QtWidgets.QVBoxLayout()
        self.wlayout.addLayout(self.endpointLayout)
        self.wlayout.addWidget(self.scanWidget)
        self.wlayout.addWidget(self.buttonBox)
        self.setLayout(self.wlayout)

    def clearLayout(self, layout):
        while layout.count():
            child = layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()

    def show_me(self):
        self.show()

    def demand_selection_of_endpoint(self, future: QtFuture):
        self.select_endpoint_future = future

    def on_endpoint_clicked(self, item: FaktsEndpoint):
        self.select_endpoint_future.resolve(item)

    def on_reject(self):
        if self.select_endpoint_future:
            self.select_endpoint_future.reject(
                Exception("User cancelled the this Grant without selecting a Beacon")
            )
        self.reject()

    def closeEvent(self, event):
        # do stuff
        if self.select_endpoint_future:
            self.select_endpoint_future.reject(
                Exception("User cancelled the this Grant without selecting a Beacon")
            )

        event.accept()  # let the window close

    def on_new_endpoint(self, config: FaktsEndpoint):
        self.clearLayout(self.endpointLayout)

        self.endpoints.append(config)

        for endpoint in self.endpoints:
            widget = FaktsEndpointWidget(endpoint)
        
            self.endpointLayout.addWidget(widget)
            widget.accept_clicked.connect(self.on_endpoint_clicked)



class QtSelectableDiscovery(AdvertisedDiscovery):
    widget: SelectBeaconWidget = Field(default_factory=SelectBeaconWidget, exclude=True)
    scan_localhost: bool = True


    async def emit_endpoints(self):

        if self.scan_localhost:
            localhost = "http://localhost:8000"
            endpoint = await self.check_beacon(localhost)
            self.widget.new_endpoint.emit(endpoint)

        try:
            async for endpoint in self.astream():
                self.widget.new_endpoint.emit(endpoint)
        except Exception as e:
            logger.exception(e)
            raise e

    async def discover(self, force_refresh=False, **kwargs):
        print("Starting Discovery")
        emitting_task = asyncio.create_task(self.emit_endpoints())
        try:

            await self.widget.show_coro.acall()
            try:
                endpoint = await self.widget.select_endpoint.acall()
                print("Endpoint selected", endpoint)
                await self.widget.hide_coro.acall()

            finally:
                emitting_task.cancel()
                try:
                    await emitting_task
                except asyncio.CancelledError as e:
                    logger.info("Cancelled the Discovery task")

            return endpoint
        except Exception as e:
            logger.exception(e)
            emitting_task.cancel()

            try:
                await emitting_task
            except asyncio.CancelledError:
                logger.info("Cancelled the Discovery task")

            raise e

    class Config:
        arbitrary_types_allowed = True



from fakts.discovery.base import Discovery
from fakts.discovery.base import FaktsEndpoint


class StaticDiscovery(Discovery):
    base_url = "http://localhost:8000/f/"

    async def discover(self):
        return FaktsEndpoint(base_url=self.base_url)

    class Config:
        extra = "forbid"



from socket import socket
from fakts.discovery.base import Discovery
from fakts.discovery.base import FaktsEndpoint
from typing import Dict, Optional

from pydantic import Field
from socket import socket, AF_INET, SOCK_DGRAM
import asyncio
import json
from koil import unkoil
import logging
import os
import yaml
import pydantic
from pydantic import BaseModel
import aiohttp
import ssl
import certifi

logger = logging.getLogger(__name__)


class DiscoveryProtocol(asyncio.DatagramProtocol):
    pass

    def __init__(self, recvq) -> None:
        super().__init__()
        self._recvq = recvq

    def datagram_received(self, data, addr):
        self._recvq.put_nowait((data, addr))


class AdvertisedConfig(BaseModel):
    selected_endpoint: Optional[FaktsEndpoint]


class ListenBinding(BaseModel):
    address: str = ""
    port: int = 45678
    magic_phrase: str = "beacon-fakts"


async def alisten(bind: ListenBinding, strict: bool = False):

    s = socket(AF_INET, SOCK_DGRAM)  # create UDP socket
    s.bind((bind.address, bind.port))

    try:

        loop = asyncio.get_event_loop()
        read_queue = asyncio.Queue()
        transport, pr = await loop.create_datagram_endpoint(
            lambda: DiscoveryProtocol(read_queue), sock=s
        )

        while True:
            data, addr = await read_queue.get()
            try:
                data = str(data, "utf8")
                if data.startswith(bind.magic_phrase):
                    endpoint = data[len(bind.magic_phrase) :]

                    try:
                        endpoint = json.loads(endpoint)
                        endpoint = FaktsEndpoint(**endpoint)
                        yield endpoint

                    except json.JSONDecodeError as e:

                        logger.error("Received Request but it was not valid json")
                        if strict:
                            raise e

                else:
                    logger.error(
                        f"Received Non Magic Response {data}. Maybe somebody sends"
                    )

            except UnicodeEncodeError as e:
                logger.error("Couldn't decode received message")
                if strict:
                    raise e

    except asyncio.CancelledError as e:
        transport.close()
        s.close()
        logger.info("Stopped checking")
        raise e
    finally:
        transport.close()
        s.close()
        logger.info("Stopped checking")


class AdvertisedDiscovery(Discovery):
    broadcast_port = 45678
    magic_phrase = "beacon-fakts"
    bind = ""
    strict: bool = False
    discovered_endpoints: Dict[str, FaktsEndpoint] = Field(default_factory=dict)
    file = ".fakts.yaml"

    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""


    async def check_beacon(self, url: str) -> FaktsEndpoint:
        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
        ) as session:
            async with session.get(
                f"{url}/.well-known/fakts",
            ) as resp:
                data = await resp.json()

                if resp.status == 200:
                    data = await resp.json()
                    if not "name" in data:
                        logger.error(f"Malformed answer: {data}")
                        raise Exception("Malformed Answer")

                    return FaktsEndpoint(**data)

                else:
                    raise Exception("Error! Coud not retrieve on the endpoint")

    async def discover(self, force_refresh=False, **kwargs):
        if os.path.exists(self.file):
            with open(self.file, "r") as f:
                x = yaml.load(f, Loader=yaml.FullLoader)
                try:
                    cache = AdvertisedConfig(**x)
                except pydantic.ValidationError as e:
                    logger.error(f"Could not load cache file: {e}. Ignoring it")
                    cache = AdvertisedConfig()
        else:
            cache = AdvertisedConfig()

        if not cache.selected_endpoint or force_refresh:
            async for i in self.astream():
                endpoint = i
                break

        with open(self.file, "w") as f:
            yaml.dump(cache.dict(), f)

        return endpoint

    async def astream(self, name_filter=None, **kwargs):

        binding = ListenBinding(address=self.bind, port=self.broadcast_port, magic_phrase=self.magic_phrase)
        async for i in alisten(binding, strict=self.strict):
            yield i

       
    def scan(self, **kwargs):
        return unkoil(self.ascan(**kwargs))

    class Config:
        arbitrary_types_allowed = True


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fakts',
 'fakts.cli',
 'fakts.discovery',
 'fakts.discovery.beacon',
 'fakts.discovery.qt',
 'fakts.fakt',
 'fakts.grants',
 'fakts.grants.io',
 'fakts.grants.io.qt',
 'fakts.grants.meta',
 'fakts.grants.remote']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.2', 'QtPy>=2.0.1,<3.0.0', 'koil>=0.2.10', 'pydantic>1.8.2']

extras_require = \
{'remote': ['aiohttp>=3.8.2,<4.0.0', 'certifi>2021']}

setup_kwargs = {
    'name': 'fakts',
    'version': '0.3.10',
    'description': 'asynchronous configuration provider ( tailored to support dynamic client-server relations)',
    'long_description': '# fakts\n\n[![codecov](https://codecov.io/gh/jhnnsrs/fakts/branch/master/graph/badge.svg?token=UGXEA2THBV)](https://codecov.io/gh/jhnnsrs/fakts)\n[![PyPI version](https://badge.fury.io/py/fakts.svg)](https://pypi.org/project/fakts/)\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg)](https://pypi.org/project/fakts/)\n![Maintainer](https://img.shields.io/badge/maintainer-jhnnsrs-blue)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/fakts.svg)](https://pypi.python.org/pypi/fakts/)\n[![PyPI status](https://img.shields.io/pypi/status/fakts.svg)](https://pypi.python.org/pypi/fakts/)\n[![PyPI download day](https://img.shields.io/pypi/dm/fakts.svg)](https://pypi.python.org/pypi/fakts/)\n\n### DEVELOPMENT\n\n## Inspiration\n\nFakts was designed to make configuration of apps compatible with concurrency pattern, it is designed to allow\nfor asynchronous retrieval of configuration from various sources, may it be a config file, environmental variables\nor a remote server.\n\n# Core Design\n\nFakts uses Grants to obtain configuration asynchronously, a grant is a way of retrieving the configuration from a\nspecific source. It can be a local config file (eg. yaml, toml, json), environemnt variables, a remote configuration (eg. from a fakts server), a database.\nThe fakts class then wraps the grant to ensure both a sychronous and asychronous interface that is threadsafe.\n\nGrants are designed to be composable through MetaGrants so by desigining a specifc grant structure, one can\nhighly customize the retrieval logic. Please check out the\n\n# Example:\n\n```python\nasync with Fakts(grant=YamlGrant("config.yaml")) as fakts:\n    config = await fakts.aget("group_name")\n```\n\nor\n\n```python\nwith Fakts(grant=YamlGrant("config.yaml")) as fakts:\n    config = fakts.get("group_name")\n```\n\nFakts should be used as a context manager, and will set the current fakts context variable to itself, letting\nyou access the current fakts instance from anywhere in your code (async or sync) without specifically passing a referece.\nTo understand how the async sync code access work, please check out the documentation for koil.\n\n# Composability\n\nYou can compose grants through meta grants in order to load configuration from multiple sources (eg. a local, file\nthat can be overwritten by a remote configuration, or some envionment variables).\n\nExample:\n\n```python\nasync with Fakts(grant=FailsafeGrant(\n    grants=[\n        EnvGrant(),\n        YamlGrant("config.yaml")\n    ]\n)) as fakts:\n    config = await fakts.get("group_name")\n```\n\nIn this example fakts will load the configuration from the environment variables first, and if that fails,\nit will load it from the yaml file.\n\n## Special Use Case: Dynamic Server Relations\n\nFakts provides the remote grant protocol for retrieval of configuration in dynamic client-server relationships.\nWith these grants you provide a software manifest for a configuration server (fakts-server), that then grants\nthe configuration (either through user approval (similar to device code grant)). These grants are mainly used\nto setup or claim an oauth2 application on the backend securely that then can be used to identify the application in the\nfuture. These grants are at the moment highly specific to the arkitekt platform and subject to change.\n\n# Sister packages\n\nThese packages provide contrib modules to support auto\nconfiguration through a fakts instance\n\n- herre: oauth2 client\n- rath: graphql client (typed through turms)\n',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from .beacon import (
    retrieve_bindings,
    advertise,
    BeaconProtocol,
    Binding,
    EndpointBeacon,
)


__all__ = [
    "retrieve_bindings",
    "advertise",
    "BeaconProtocol",
    "Binding",
    "EndpointBeacon",
]



import collections.abc
from typing import Mapping


def update_nested(d: Mapping, u: Mapping) -> Mapping:
    """Update a nested dictionary or similar mapping.
    This is a recursive function that will update the values in the dictionary
    *inplace*.

    Args:
        d (Mapping): The dictionary to update
        u (Mapping): The dictionary to update with

    Returns:
        Mapping: The updated dictionary
    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_nested(d.get(k, {}), v)
        else:
            d[k] = v
    return d
